!---------------------------------------------------------------------------
!
! This module reads the bufr file for different instruments for OSSE Package
! 
! This Module adds random noise to the simulated radiances 
! 
! CREATION HISTORY:
!---------------------------------------------------------------------------
! Created | Narges Shahroudi   | narges.shahroudi@noaa.gov | 01-May-2015 



MODULE Read_Data_Module

  USE CRTM_Module
   
  IMPLICIT NONE
    
  TYPE ::  BufrObservation_type
  
     Character(30)       :: satid=''
     INTEGER             :: nProf =0
     INTEGER,ALLOCATABLE :: itype(:)           
     INTEGER,ALLOCATABLE :: Chan (:)                    
     REAL(8),ALLOCATABLE :: year(:,:)             
     REAL(8),ALLOCATABLE :: month (:,:)           
     REAL(8),ALLOCATABLE :: day (:,:)             
     REAL(8),ALLOCATABLE :: hour(:,:)             
     REAL(8),ALLOCATABLE :: minute(:,:)           
     REAL(8),ALLOCATABLE :: second(:,:)           
     REAL(8),ALLOCATABLE :: lat(:,:)              
     REAL(8),ALLOCATABLE :: lon(:,:)              
     REAL(8),ALLOCATABLE :: soza(:,:)             
     REAL(8),ALLOCATABLE :: FOV(:,:)              
     
  END TYPE BufrObservation_type


CONTAINS

  SUBROUTINE Read_Bufr_File(fname_inbufr,nchan,Sensor_ID_IN,BufrInfo,Observation)  


  CHARACTER(*), INTENT(IN) ::  fname_inbufr
  CHARACTER(*), INTENT(IN) ::  Sensor_ID_IN
  INTEGER, PARAMETER :: lnbufr = 10
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  INTEGER nchan
  CHARACTER(LEN=100) :: Sensor_ID



  OPEN(UNIT=lnbufr, FILE=fname_inbufr, FORM="UNFORMATTED")
  PRINT*,'Reading: ',fname_inbufr 

 Sensor_ID=TRIM(Sensor_ID_IN)



  IF (INDEX(Sensor_ID, "aqua") > 0) THEN
      Sensor_ID = "aqua"
  ENDIF
 

    

  IF     (INDEX(Sensor_ID, "amsu") > 0 .OR.&
          INDEX(Sensor_ID, "hirs") > 0 )THEN
          
          
          CALL Read_TOVS_Bufr(lnbufr,nchan,BufrInfo,Observation)  
         
 ELSEIF (INDEX(Sensor_ID, "atms") > 0 )THEN

           CALL Read_ATMS_Bufr(lnbufr,nchan,BufrInfo,Observation)  

ELSEIF (INDEX(Sensor_ID, "mhs") > 0 )THEN

           CALL Read_MHS_Bufr(lnbufr,nchan,BufrInfo,Observation)  

 ELSEIF  (INDEX(Sensor_ID, "aqua") > 0 ) THEN  
          
           CALL Read_AIRS_Bufr(lnbufr,nchan,BufrInfo,Observation)
         
 ELSEIF  (INDEX(Sensor_ID, "iasi") > 0 )  THEN

          CALL Read_IASI_Bufr(lnbufr,nchan,BufrInfo,Observation)

 ELSEIF  (INDEX(Sensor_ID, "sndr") > 0 )  THEN
       
          CALL Read_GOESNDR_Bufr(lnbufr,nchan,BufrInfo,Observation)

 ELSEIF  (INDEX(Sensor_ID, "cris") > 0 )  THEN

          CALL Read_CRIS_Bufr(lnbufr,nchan,BufrInfo,Observation)

 ELSEIF  (INDEX(Sensor_ID, "seviri") > 0 )  THEN
          
          CALL Read_SEVIRI_Bufr(lnbufr,nchan,BufrInfo,Observation)

 ELSEIF  (INDEX(Sensor_ID, "ssmis") > 0 )  THEN
          
          CALL Read_SSMIS_Bufr(lnbufr,nchan,BufrInfo,Observation)
          
 ENDIF
    
 
 Close (lnbufr)
 
END SUBROUTINE Read_Bufr_File


SUBROUTINE Read_TOVS_Bufr(lnbufr,nchan,BufrInfo,Observation)
  

  INTEGER, INTENT(IN) ::  lnbufr,nchan 
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:) 
  REAL(8), ALLOCATABLE :: TB(:)
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLATH CLONH SAZA SOZA HMSL LSQL'/
  
  nnum=0

      ALLOCATE (TB(nchan))
      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)      
        
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 .and. nnum <  mxtb)
       
            CALL ufbint(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)

             nnum=nnum+1        
             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)

              CALL ufbrep(lnbufr,TB,1,nchan,iret,'TMBR')
              TB2(1:nchan,nnum)=TB(1:nchan)
            

           ENDDO
        ENDDO bufrobs


        ALLOCATE (Observation(nchan,nnum))
        ALLOCATE (BufrInfo%lat(nnum,1))
        ALLOCATE (BufrInfo%lon(nnum,1))
        ALLOCATE (BufrInfo%soza(nnum,1))
        ALLOCATE (BufrInfo%FOV(nnum,1))
        ALLOCATE (BufrInfo%year(nnum,1))
        ALLOCATE (BufrInfo%month(nnum,1))
        ALLOCATE (BufrInfo%day(nnum,1))
        ALLOCATE (BufrInfo%hour(nnum,1))
        ALLOCATE (BufrInfo%minute(nnum,1))
        ALLOCATE (BufrInfo%second(nnum,1)) 
        ALLOCATE (BufrInfo%itype(1))
        ALLOCATE (BufrInfo%chan(1)) 
     
        BufrInfo %itype=itype1  
        BufrInfo%nprof=nnum
        BufrInfo%chan=nchan   
        BufrInfo%FOV(:,1)=FOV1(1:nnum)
        BufrInfo%year(:,1)=year1(1:nnum)
        BufrInfo%month(:,1)=month1(1:nnum)
        BufrInfo%day(:,1)=day1(1:nnum)
        BufrInfo%hour(:,1)=hour1(1:nnum)
        BufrInfo%minute(:,1)=minute1(1:nnum)
        BufrInfo%second(:,1)=second1(1:nnum)  
        BufrInfo%soza(:,1)=soza1(1:nnum)
        BufrInfo%lat(:,1)=lat1(1:nnum)
        BufrInfo%lon(:,1)=lon1(1:nnum)
        Observation%Brightness_Temperature=TB2(1:nchan,1:nnum)
       
        
           
        CALL closbf(lnbufr)
   
END SUBROUTINE Read_TOVS_Bufr

 

SUBROUTINE Read_AIRS_Bufr(lnbufr,nchan,BufrInfo,Observation)
  
  INTEGER, INTENT(IN) ::  lnbufr,nchan

  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  REAL(8), ALLOCATABLE :: TB(:) 
  REAL(8),dimension(14,2):: bfr1bhdr
  REAL(8) bfh(5)
  CHARACTER(30) satid,str
  CHARACTER(8) subset
  CHARACTER (80) hdr1b  
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1(2),i,irec,isub,airs_ch
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=14,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb,2),lon1(mxtb,2),soza1(mxtb,2),FOV1(mxtb,2),year1(mxtb,2),month1(mxtb,2),day1(mxtb,2),hour1(mxtb,2),minute1(mxtb,2),second1(mxtb,2)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SIID FOVN YEAR MNTH DAYS HOUR MINU SECO CLATH CLONH SAZA SOZA HMSL LSQL'/
  
  nnum=0
  airs_ch=301

      ALLOCATE (TB(airs_ch))
    
      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)
          
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)>=0 )
     
        read_loop: DO WHILE (ireadsb(lnbufr)==0 )
                    
           CALL ufbrep(lnbufr,bfr1bhdr,n1bhdr,2,iret,hdr1b)
          
          nnum=nnum+1      
                 
            DO i=1,2
               itype1(i)=bfr1bhdr(1,i)  
               FOV1(nnum,i)=bfr1bhdr(2,i)
               year1(nnum,i)=bfr1bhdr(3,i)
               month1(nnum,i)=bfr1bhdr(4,i)
               day1(nnum,i)=bfr1bhdr(5,i)  
               hour1(nnum,i)=bfr1bhdr(6,i)
               minute1(nnum,i)=bfr1bhdr(7,i)
               second1(nnum,i)=bfr1bhdr(8,i)          
               soza1(nnum,i)=bfr1bhdr(12,i)
               lat1(nnum,i)=bfr1bhdr(9,i)
               lon1(nnum,i)=bfr1bhdr(10,i)
            ENDDO            
                             
           CALL ufbrep(lnbufr,TB,1,airs_ch,iret,'TMBR')  
          
             TB2(1:airs_ch,nnum)=TB(1:airs_ch)
             
         ENDDO read_loop
        
   ENDDO bufrobs


   ALLOCATE (Observation(airs_ch,nnum ))
   ALLOCATE (BufrInfo%lat(nnum,2))
   ALLOCATE (BufrInfo%lon(nnum,2))
   ALLOCATE (BufrInfo%soza(nnum,2))
   ALLOCATE (BufrInfo%FOV(nnum,2))
   ALLOCATE (BufrInfo%year(nnum,2))
   ALLOCATE (BufrInfo%month(nnum,2))
   ALLOCATE (BufrInfo%day(nnum,2))
   ALLOCATE (BufrInfo%hour(nnum,2))
   ALLOCATE (BufrInfo%minute(nnum,2))
   ALLOCATE (BufrInfo%second(nnum,2)) 
   ALLOCATE (BufrInfo%itype(2))
   ALLOCATE (BufrInfo%chan(2))
                    
  
   BufrInfo %itype(1:2)=itype1(1:2)  
   BufrInfo%nprof=nnum
   BufrInfo%chan=airs_ch
   BufrInfo%FOV(:,1:2)=FOV1(1:nnum,1:2)
   BufrInfo%year(:,1:2)=year1(1:nnum,1:2)
   BufrInfo%month(:,1:2)=month1(1:nnum,1:2)
   BufrInfo%day(:,1:2)=day1(1:nnum,1:2)
   BufrInfo%hour(:,1:2)=hour1(1:nnum,1:2)
   BufrInfo%minute(:,1:2)=minute1(1:nnum,1:2)
   BufrInfo%second(:,1:2)=second1(1:nnum,1:2)  
   BufrInfo%soza(:,1:2)=soza1(1:nnum,1:2)
   BufrInfo%lat(:,1:2)=lat1(1:nnum,1:2)
   BufrInfo%lon(:,1:2)=lon1(1:nnum,1:2)  
   Observation%Brightness_Temperature=TB2(1:airs_ch,1:nnum)
   
   
   CALL closbf(lnbufr)
 
   
END SUBROUTINE Read_AIRS_Bufr

SUBROUTINE  Read_IASI_Bufr(lnbufr,nchan,BufrInfo,Observation)
   
  INTEGER, INTENT(IN) ::  lnbufr,nchan
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  REAL(8), ALLOCATABLE :: TB(:,:) 
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid,str
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1,i,j
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:),CHMN(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLATH CLONH SAZA SOZA HMSL LSQL'/
  real(8)::    scal(3,10)
  
  nnum=0

      ALLOCATE (TB(2,nchan))
      ALLOCATE (CHMN(nchan))

      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)
     
  
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 .and. nnum <  mxtb)
            
            CALL ufbint(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)
            
            CALL ufbrep(lnbufr,scal,3,10,iret,'STCH ENCH CHSF')
      
             nnum=nnum+1          
             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)         
           
             CALL ufbrep(lnbufr,TB,2,nchan,iret,'CHNM SCRA')
                       
             CHMN(1:nchan)=TB(1,1:nchan)
             TB2(1:nchan,nnum)=TB(2,1:nchan)
             
             DO i=1,nchan
               DO j=1,10 
                  IF (CHMN(i)>=scal(1,j).and.CHMN(i)<=scal(2,j))THEN                
                     TB2(i,nnum)=TB2(i,nnum)/(10**(scal(3,j)-5))        
                  ENDIF
               ENDDO
            ENDDO
         ENDDO
      ENDDO bufrobs
          

   ALLOCATE( Observation(nchan,nnum ) )
   ALLOCATE (BufrInfo%lat(nnum,1))
   ALLOCATE (BufrInfo%lon(nnum,1))
   ALLOCATE (BufrInfo%soza(nnum,1))
   ALLOCATE (BufrInfo%FOV(nnum,1))
   ALLOCATE (BufrInfo%year(nnum,1))
   ALLOCATE (BufrInfo%month(nnum,1))
   ALLOCATE (BufrInfo%day(nnum,1))
   ALLOCATE (BufrInfo%hour(nnum,1))
   ALLOCATE (BufrInfo%minute(nnum,1))
   ALLOCATE (BufrInfo%second(nnum,1))    
   ALLOCATE (BufrInfo%itype(1))
   ALLOCATE (BufrInfo%chan(1))
   
   
   BufrInfo %itype=itype1  
   BufrInfo%nprof=nnum
   BufrInfo%chan=nchan
   BufrInfo%FOV(:,1)=FOV1(1:nnum)
   BufrInfo%year(:,1)=year1(1:nnum)
   BufrInfo%month(:,1)=month1(1:nnum)
   BufrInfo%day(:,1)=day1(1:nnum)
   BufrInfo%hour(:,1)=hour1(1:nnum)
   BufrInfo%minute(:,1)=minute1(1:nnum)
   BufrInfo%second(:,1)=second1(1:nnum)  
   BufrInfo%soza(:,1)=soza1(1:nnum)
   BufrInfo%lat(:,1)=lat1(1:nnum)
   BufrInfo%lon(:,1)=lon1(1:nnum)
   
   Observation%Radiance=TB2(1:nchan,1:nnum)

   CALL closbf(lnbufr)
   
 END SUBROUTINE Read_IASI_Bufr


 SUBROUTINE  Read_GOESNDR_Bufr(lnbufr,nchan,BufrInfo,Observation)

  INTEGER, INTENT(IN) ::  lnbufr,nchan
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  REAL(8), ALLOCATABLE :: TB(:) 
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid,str
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLAT CLON SAZA SOZA HMSL LSQL'/
  
  nnum=0
  
      ALLOCATE (TB(nchan))
      
      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)    
   
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 .and. nnum <  mxtb)
       
            CALL ufbint(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)
             nnum=nnum+1
           
             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)
             
             CALL ufbrep(lnbufr,TB,1,nchan,iret,'TMBR')
            
             TB2(1:nchan,nnum)=TB(1:nchan)

          ENDDO
       ENDDO bufrobs
      

   ALLOCATE( Observation(nchan,nnum ) )
   
   ALLOCATE (BufrInfo%lat(nnum,1))
   ALLOCATE (BufrInfo%lon(nnum,1))
   ALLOCATE (BufrInfo%soza(nnum,1))
   ALLOCATE (BufrInfo%FOV(nnum,1))
   ALLOCATE (BufrInfo%year(nnum,1))
   ALLOCATE (BufrInfo%month(nnum,1))
   ALLOCATE (BufrInfo%day(nnum,1))
   ALLOCATE (BufrInfo%hour(nnum,1))
   ALLOCATE (BufrInfo%minute(nnum,1))
   ALLOCATE (BufrInfo%second(nnum,1)) 
   ALLOCATE (BufrInfo%itype(1))
   ALLOCATE (BufrInfo%chan(1))
   
  
   BufrInfo %itype=itype1  
   BufrInfo%nprof=nnum
   BufrInfo%chan=nchan
   BufrInfo%FOV(:,1)=FOV1(1:nnum)
   BufrInfo%year(:,1)=year1(1:nnum)
   BufrInfo%month(:,1)=month1(1:nnum)
   BufrInfo%day(:,1)=day1(1:nnum)
   BufrInfo%hour(:,1)=hour1(1:nnum)
   BufrInfo%minute(:,1)=minute1(1:nnum)
   BufrInfo%second(:,1)=second1(1:nnum)  
   BufrInfo%soza(:,1)=soza1(1:nnum)
   BufrInfo%lat(:,1)=lat1(1:nnum)
   BufrInfo%lon(:,1)=lon1(1:nnum)
  
   Observation%Brightness_Temperature=TB2(1:nchan,1:nnum)

   CALL closbf(lnbufr)
   
   
 END SUBROUTINE Read_GOESNDR_Bufr
 
 SUBROUTINE  Read_SSMIS_Bufr(lnbufr,nchan,BufrInfo,Observation)
  

  INTEGER, INTENT(IN) ::  lnbufr,nchan
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  REAL(8), ALLOCATABLE :: TB(:)
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid,str
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLATH CLONH SAZA SOZA HMSL LSQL'/
 
  nnum=0

      ALLOCATE (TB(nchan))
    
      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)
       
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 )
       
            CALL ufbint(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)
             nnum=nnum+1          
             
             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)

           
             CALL ufbrep(lnbufr,TB,1,nchan,iret,'TMBR')
             TB2(1:nchan,nnum)=TB(1:nchan)         

      ENDDO
   ENDDO bufrobs
   
   ALLOCATE( Observation(nchan,nnum ) )

   ALLOCATE (BufrInfo%lat(nnum,1))
   ALLOCATE (BufrInfo%lon(nnum,1))
   ALLOCATE (BufrInfo%soza(nnum,1))
   ALLOCATE (BufrInfo%FOV(nnum,1))
   ALLOCATE (BufrInfo%year(nnum,1))
   ALLOCATE (BufrInfo%month(nnum,1))
   ALLOCATE (BufrInfo%day(nnum,1))
   ALLOCATE (BufrInfo%hour(nnum,1))
   ALLOCATE (BufrInfo%minute(nnum,1))
   ALLOCATE (BufrInfo%second(nnum,1))  
   ALLOCATE (BufrInfo%itype(1))
   ALLOCATE (BufrInfo%chan(1))
   
  
   BufrInfo %itype=itype1  
   BufrInfo%nprof=nnum
   BufrInfo%chan=nchan
   BufrInfo%FOV(:,1)=FOV1(1:nnum)
   BufrInfo%year(:,1)=year1(1:nnum)
   BufrInfo%month(:,1)=month1(1:nnum)
   BufrInfo%day(:,1)=day1(1:nnum)
   BufrInfo%hour(:,1)=hour1(1:nnum)
   BufrInfo%minute(:,1)=minute1(1:nnum)
   BufrInfo%second(:,1)=second1(1:nnum)  
   BufrInfo%soza(:,1)=soza1(1:nnum)
   BufrInfo%lat(:,1)=lat1(1:nnum)
   BufrInfo%lon(:,1)=lon1(1:nnum)
   
   Observation%Brightness_Temperature=TB2(1:nchan,1:nnum)

   CALL closbf(lnbufr)
   
 END SUBROUTINE Read_SSMIS_Bufr


 SUBROUTINE  Read_CRIS_Bufr(lnbufr,nchan,BufrInfo,Observation)
   
  INTEGER, INTENT(IN) ::  lnbufr,nchan

  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  REAL(8), ALLOCATABLE :: TB(:) 
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid,str
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLATH CLONH SAZA SOZA HMSL LSQL'/

  nnum=0


      ALLOCATE (TB(nchan))
    
      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)
      CALL readmg(lnbufr,subset,idate,iret) 
      
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 .and. nnum <  mxtb)
       
            CALL ufbint(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)
             nnum=nnum+1
             

             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)
             
             CALL ufbrep(lnbufr,TB,1,nchan,iret,'SRAD')
             TB2(1:nchan,nnum)=TB(1:nchan)*1000
             
          ENDDO
       ENDDO bufrobs
 
       ALLOCATE( Observation(nchan,nnum ))
       ALLOCATE (BufrInfo%lat(nnum,1))
       ALLOCATE (BufrInfo%lon(nnum,1))
       ALLOCATE (BufrInfo%soza(nnum,1))
       ALLOCATE (BufrInfo%FOV(nnum,1))
       ALLOCATE (BufrInfo%year(nnum,1))
       ALLOCATE (BufrInfo%month(nnum,1))
       ALLOCATE (BufrInfo%day(nnum,1))
       ALLOCATE (BufrInfo%hour(nnum,1))
       ALLOCATE (BufrInfo%minute(nnum,1))
       ALLOCATE (BufrInfo%second(nnum,1))      
       ALLOCATE (BufrInfo%itype(1))
       ALLOCATE (BufrInfo%chan(1))
   
      
       BufrInfo %itype=itype1  
       BufrInfo%nprof=nnum
       BufrInfo%chan=nchan
       BufrInfo%FOV(:,1)=FOV1(1:nnum)
       BufrInfo%year(:,1)=year1(1:nnum)
       BufrInfo%month(:,1)=month1(1:nnum)
       BufrInfo%day(:,1)=day1(1:nnum)
       BufrInfo%hour(:,1)=hour1(1:nnum)
       BufrInfo%minute(:,1)=minute1(1:nnum)
       BufrInfo%second(:,1)=second1(1:nnum)  
       BufrInfo%soza(:,1)=soza1(1:nnum)
       BufrInfo%lat(:,1)=lat1(1:nnum)
       BufrInfo%lon(:,1)=lon1(1:nnum)
       
       Observation%Radiance=TB2(1:nchan,1:nnum)
       
       CALL closbf(lnbufr)
     END SUBROUTINE Read_CRIS_Bufr
     
SUBROUTINE  Read_SEVIRI_Bufr(lnbufr,nchan,BufrInfo,Observation)

  INTEGER, INTENT(IN) ::  lnbufr,nchan
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  REAL(8), ALLOCATABLE :: TB(:) 
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid,str
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1,seviri_ch
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLATH CLONH SAZA SOZA HMSL LSQL'/
 
  nnum=0
  seviri_ch=11


      ALLOCATE (TB(seviri_ch))

      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)
     
      
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 .and. nnum <  mxtb)
            
            CALL ufbint(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)
             nnum=nnum+1
             

             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)
             
             CALL ufbrep(lnbufr,TB,1,seviri_ch,iret,'TMBRST')
             TB2(1:seviri_ch,nnum)=TB(1:seviri_ch)
             
          ENDDO
       ENDDO bufrobs
    
       
       ALLOCATE( Observation(seviri_ch,nnum ))   
       ALLOCATE (BufrInfo%lat(nnum,1))
       ALLOCATE (BufrInfo%lon(nnum,1))
       ALLOCATE (BufrInfo%soza(nnum,1))
       ALLOCATE (BufrInfo%FOV(nnum,1))
       ALLOCATE (BufrInfo%year(nnum,1))
       ALLOCATE (BufrInfo%month(nnum,1))
       ALLOCATE (BufrInfo%day(nnum,1))
       ALLOCATE (BufrInfo%hour(nnum,1))
       ALLOCATE (BufrInfo%minute(nnum,1))
       ALLOCATE (BufrInfo%second(nnum,1)) 
       ALLOCATE (BufrInfo%itype(1))
       ALLOCATE (BufrInfo%chan(1))

      
       BufrInfo %itype=itype1  
       BufrInfo%nprof=nnum
       BufrInfo%chan=seviri_ch
       BufrInfo%FOV(:,1)=FOV1(1:nnum)
       BufrInfo%year(:,1)=year1(1:nnum)
       BufrInfo%month(:,1)=month1(1:nnum)
       BufrInfo%day(:,1)=day1(1:nnum)
       BufrInfo%hour(:,1)=hour1(1:nnum)
       BufrInfo%minute(:,1)=minute1(1:nnum)
       BufrInfo%second(:,1)=second1(1:nnum)  
       BufrInfo%soza(:,1)=soza1(1:nnum)
       BufrInfo%lat(:,1)=lat1(1:nnum)
       BufrInfo%lon(:,1)=lon1(1:nnum)
       
       Observation%Brightness_Temperature=TB2(1:seviri_ch,1:nnum)
       CALL closbf(lnbufr)
       
     END SUBROUTINE Read_SEVIRI_Bufr

SUBROUTINE  Read_ATMS_Bufr(lnbufr,nchan,BufrInfo,Observation)

  INTEGER, INTENT(IN) ::  lnbufr,nchan
 
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:)
  REAL(8), ALLOCATABLE :: TB(:) 
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid,str
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,  ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=20000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLATH CLONH SAZA SOZA HMSL LSQL'/
 
  nnum=0

      ALLOCATE (TB(nchan))
      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)
    
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 .and. nnum <  mxtb)
          
            CALL ufbrep(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)
             nnum=nnum+1
             
             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)
             
             CALL ufbrep(lnbufr,TB,1,nchan,iret,'TMANT')
             TB2(1:nchan,nnum)=TB(1:nchan)
             
          ENDDO
       ENDDO bufrobs

   
       ALLOCATE( Observation(nchan,nnum ) )
       
       ALLOCATE (BufrInfo%lat(nnum,1))
       ALLOCATE (BufrInfo%lon(nnum,1))
       ALLOCATE (BufrInfo%soza(nnum,1))
       ALLOCATE (BufrInfo%FOV(nnum,1))
       ALLOCATE (BufrInfo%year(nnum,1))
       ALLOCATE (BufrInfo%month(nnum,1))
       ALLOCATE (BufrInfo%day(nnum,1))
       ALLOCATE (BufrInfo%hour(nnum,1))
       ALLOCATE (BufrInfo%minute(nnum,1))
       ALLOCATE (BufrInfo%second(nnum,1))
       ALLOCATE (BufrInfo%itype(1))
       ALLOCATE (BufrInfo%chan(1))
       
      
       BufrInfo %itype=itype1  
       BufrInfo%nprof=nnum
       BufrInfo%chan=nchan
       BufrInfo%FOV(:,1)=FOV1(1:nnum)
       BufrInfo%year(:,1)=year1(1:nnum)
       BufrInfo%month(:,1)=month1(1:nnum)
       BufrInfo%day(:,1)=day1(1:nnum)
       BufrInfo%hour(:,1)=hour1(1:nnum)
       BufrInfo%minute(:,1)=minute1(1:nnum)
       BufrInfo%second(:,1)=second1(1:nnum)  
       BufrInfo%soza(:,1)=soza1(1:nnum)
       BufrInfo%lat(:,1)=lat1(1:nnum)
       BufrInfo%lon(:,1)=lon1(1:nnum)

       Observation%Brightness_Temperature=TB2(1:nchan,1:nnum)
             
                 
       CALL closbf(lnbufr)
       
END SUBROUTINE Read_ATMS_Bufr

SUBROUTINE  Read_MHS_Bufr(lnbufr,nchan,BufrInfo,Observation)

  INTEGER, INTENT(IN) ::  lnbufr,nchan 
  TYPE(BufrObservation_type) :: BufrInfo
  TYPE(CRTM_RTSolution_type), ALLOCATABLE :: Observation(:,:) 
  REAL(8), ALLOCATABLE :: TB(:)
  REAL(8),dimension(25):: bfr1bhdr
  CHARACTER(30) satid
  CHARACTER(8) subset
  CHARACTER (80) hdr1b
  INTEGER iret,idate,nnum,ireadmg,ireadsb,nprof,iunit,itype1
  REAL (8), ALLOCATABLE ::lat(:,:),lon(:,:),soza(:,:),FOV(:,:),year(:,:),month(:,:),day(:,:),hour(:,:),minute(:,:),second(:,:)
  INTEGER,ALLOCATABLE ::itype(:),Chan(:)
  INTEGER,PARAMETER:: n1bhdr=25,mxtb=10000,mchan=700
  REAL(8) lat1(mxtb),lon1(mxtb),soza1(mxtb),FOV1(mxtb),year1(mxtb),month1(mxtb),day1(mxtb),hour1(mxtb),minute1(mxtb),second1(mxtb)
  REAL(8) TB2(mchan,mxtb)
  data hdr1b /'SAID FOVN YEAR MNTH DAYS HOUR MINU SECO CLAT CLON SAZA SOZA HMSL LSQL'/
  
  nnum=0

      ALLOCATE (TB(nchan))
      CALL openbf(lnbufr,'IN',lnbufr)
      CALL datelen(10)      
        
      bufrobs: DO WHILE (ireadmg(lnbufr,subset,idate)==0 )
         DO WHILE (ireadsb(lnbufr)==0 .and. nnum <  mxtb)
       
            CALL ufbint(lnbufr,bfr1bhdr,n1bhdr,1,iret,hdr1b)
           
            
             nnum=nnum+1        
             itype1=(bfr1bhdr(1))  
             FOV1(nnum)=bfr1bhdr(2)
             year1(nnum)=bfr1bhdr(3)
             month1(nnum)=bfr1bhdr(4)
             day1(nnum)=bfr1bhdr(5)  
             hour1(nnum)=bfr1bhdr(6)
             minute1(nnum)=bfr1bhdr(7)
             second1(nnum)=bfr1bhdr(8)          
             soza1(nnum)=bfr1bhdr(12)
             lat1(nnum)=bfr1bhdr(9)
             lon1(nnum)=bfr1bhdr(10)

              CALL ufbrep(lnbufr,TB,1,nchan,iret,'TMBR')
              TB2(1:nchan,nnum)=TB(1:nchan)
         

           ENDDO
        ENDDO bufrobs
     
      
   
       ALLOCATE( Observation(nchan,nnum ) )
    
       ALLOCATE (BufrInfo%lat(nnum,1))
       ALLOCATE (BufrInfo%lon(nnum,1))
       ALLOCATE (BufrInfo%soza(nnum,1))
       ALLOCATE (BufrInfo%FOV(nnum,1))
       ALLOCATE (BufrInfo%year(nnum,1))
       ALLOCATE (BufrInfo%month(nnum,1))
       ALLOCATE (BufrInfo%day(nnum,1))
       ALLOCATE (BufrInfo%hour(nnum,1))
       ALLOCATE (BufrInfo%minute(nnum,1))
       ALLOCATE (BufrInfo%second(nnum,1))
       ALLOCATE (BufrInfo%itype(1))
       ALLOCATE (BufrInfo%chan(1))
       
      
       BufrInfo %itype=itype1  
       BufrInfo%nprof=nnum
       BufrInfo%chan=nchan
       BufrInfo%FOV(:,1)=FOV1(1:nnum)
       BufrInfo%year(:,1)=year1(1:nnum)
       BufrInfo%month(:,1)=month1(1:nnum)
       BufrInfo%day(:,1)=day1(1:nnum)
       BufrInfo%hour(:,1)=hour1(1:nnum)
       BufrInfo%minute(:,1)=minute1(1:nnum)
       BufrInfo%second(:,1)=second1(1:nnum)  
       BufrInfo%soza(:,1)=soza1(1:nnum)
       BufrInfo%lat(:,1)=lat1(1:nnum)
       BufrInfo%lon(:,1)=lon1(1:nnum)

       Observation%Brightness_Temperature=TB2(1:nchan,1:nnum)
       print*,BufrInfo%lat(:,1)
             
       
       CALL closbf(lnbufr)

     END SUBROUTINE Read_MHS_Bufr


END MODULE READ_DATA_MODULE









   
  
