!---------------------------------------------------------------------------
!
! Add random noise  for OSSE Package
! 
! This Module adds random noise to the satellite radiance 
! 
! CREATION HISTORY:
!---------------------------------------------------------------------------
! Created | Narges Shahroudi   | narges.shahroudi@noaa.gov | 01-May-2015 |
!---------------------------------------------------------------------------


MODULE Sim_Obs_Module
  USE m_Kinds
  USE Read_data_Module
  USE CRTM_Module
  USE Physics_Module
  USE SpcCoeff_Define 

  IMPLICIT NONE
  



Contains

  SUBROUTINE Add_Obs_Err(BufrInfo,Observation,spccoeff,Observation_bias)
  Real  year,month,day
  REAL(8) TB2,Rad,Rad2,TB3,bmiss
  INTEGER nprof,nchan,i,n
  CHARACTER(80) idsensor,satid
  TYPE(SpcCoeff_type) :: spccoeff
  TYPE(BufrObservation_type), INTENT (IN):: BufrInfo
  TYPE(CRTM_RTSolution_type),  INTENT(IN) :: Observation(:,:)
  TYPE(CRTM_RTSolution_type),ALLOCATABLE :: Observation_bias(:,:)
  REAL(8),ALLOCATABLE :: TB(:)
  INTEGER             :: chidx
  
  
 
    chidx=maxval(spccoeff%Sensor_Channel)
    
    nchan=BufrInfo%chan(1)
    

    satid=spccoeff%Sensor_Id
    nprof=BufrInfo%nprof
    bmiss=99999.99
  
    ALLOCATE (Observation_bias(nchan,nprof ))
    ALLOCATE (TB(nchan))

    DO i=1,nprof
 
    
       IF (INDEX(satid,'iasi')>0.OR.INDEX(satid,'cris')>0)THEN
          
          
          DO n=1,nchan

             Rad=Observation(n,i)%Radiance
                      
         !    IF (Rad>bmiss)THEN
         !       TB(n)=bmiss    
         !    ELSE
                CALL Planck_Temperature( spccoeff, n , Rad, TB2 )
                TB(n)=TB2
        !     ENDIF            
             
           
            
          ENDDO
          
               
      ELSE

         TB(1:nchan)=Observation(:,i)%Brightness_Temperature      

      ENDIF
      
     
     


      CALL Add_noise(satid,nchan,chidx,TB,nprof)
   
    
      IF (INDEX(satid,'iasi')>0.OR.INDEX(satid,'cris')>0)THEN
         

         DO n=1,nchan
            
            TB3=TB(n)

           
           
          !  IF (TB3>bmiss)THEN
          !     observation_bias(n,i)%Radiance=bmiss
          !  ELSE
               CALL Planck_Radiance(spccoeff,n , TB3, Rad2) 
               Observation_bias(n,i)%Radiance=Rad2
          !  ENDIF
       
            ENDDO
     
      
      ELSE
         
         Observation_bias(1:nchan,i)%Brightness_Temperature=TB(1:nchan)

      ENDIF
      
   ENDDO
 
  
   PRINT*,'Adding random noise to:',satid
 
   
 END SUBROUTINE Add_Obs_Err
 

SUBROUTINE Add_noise(satid,nchan,chidx,TB,nprof)

      INTEGER :: j,chidx,nprof,nchan,i,count
      CHARACTER(len=80) :: header,sat(240),satid
      REAL(8),ALLOCATABLE :: var(:),stdv(:),TB(:) 
      INTEGER channel(240)
      REAL(8) variance(240)
      REAL(8) pert
      INTEGER, PARAMETER :: lun = 30
      
    
      open(lun,file='rad_varaddition.txt', form='formatted')
   
      ALLOCATE(stdv(nchan))
      count=0
      stdv=0
      
      read (lun,'(a30) ')header
   
          do j=1,239
        
             read(lun,*)sat(j),channel(j),variance(j)
            
          
             IF (satid==sat(j))THEN

                   count=count+1
                  
                IF (channel(j)<=chidx)THEN
               
                    
                   IF (nchan /=chidx) THEN
                   
                      stdv(count)=sqrt(variance(j))
                   
                   ElSE
                      
                      stdv(channel(j))=sqrt(variance(j))
                      
                   ENDIF
                   
                ENDIF
                
             ENDIF
             
          ENDDO
      
          
         DO i=1,nchan
            
            call pert_gauss (zero_k3,stdv(i),pert)
            
            TB(i)=TB(i)+pert
            
         ENDDO
       
        
            
        
      CLOSE(lun)

END SUBROUTINE Add_noise


 SUBROUTINE pert_gauss (mean,stdv,draw1)
    
     integer, parameter :: ndraw=12
     real(rkind3), intent(in)   :: mean
     real(rkind3), intent(in)   :: stdv
     real(rkind3), intent(out)  :: draw1
     integer      :: n
     real(rkind3) :: x, xsum
  
     xsum=zero_k3
    
     DO n=1,ndraw
        
        CALL random_number (x)
        xsum=xsum+x
       
     ENDDO
    
     x=(xsum-0.5_rkind3*ndraw)*sqrt(12._rkind3/ndraw)
     draw1=x*stdv+mean
    
   END SUBROUTINE pert_gauss

 
 END MODULE Sim_Obs_Module

